#include <cstddef>
#include <cstdlib>
#include <iostream>
#include <vector>
#include <list>
#include <bitset>
#include <unistd.h>
#include <iomanip>
#include <sstream>
#include <cmath>
#include <algorithm>
#include <err.h>

using namespace std;

bool flag_quiet = false;

const size_t max_value = 2000; 

typedef bitset<max_value> factors;
typedef vector<factors>   factors_matrix;

template<class COLLECTION>
inline
std::string
join_collection(const COLLECTION &col, const std::string& delimiter="\t")
{
	std::ostringstream os;
	bool first = true ;

	typename COLLECTION::const_iterator it = col.begin();

	while ( it != col.end() ) {
		if ( first ) 
			first = false;
		else
			os << delimiter ;
		os << *it ;
		it++ ;
	}

	return os.str();
}


/*
 * returns a bitset of all the factors for the given value.
 *
 * notes:
 *      bitset starts at value ZERO, which of course is never a valid factor, but this
 *      makes the rest of the processing easier (as other arrays starts at 0).
 *
 * 	1 is taken for granted and NEVER set in the returned bitset.
 *
 * 	If the given value is prime, it is set in the returned bitset.
 */
factors make_factorization_bitset ( const size_t value )
{
	factors results;
	size_t i = 2;
	size_t temp = value ;
	while ( i <= (value/2) ) {
		if ( temp == 1 )
			break ;
		if ( (temp%i)==0 ) {
			results.set(i);		
			temp/= i ;
		} else
			i++;
	}
	//Prime? set itself as a factor
	if (results.none())
		results.set(value);
	return results;
}

/*
 * crappy but easy way to tell if a value is prime
 */
inline bool is_prime ( const size_t value )
{
	factors f = make_factorization_bitset(value);
	return ( f.count()==1 && f.test(value) ) ;
}

factors_matrix build_factors_matrix ( size_t value )
{
	factors_matrix results;

	for ( size_t i = 0 ; i <= value ; ++i ) {
		results.push_back ( make_factorization_bitset ( i ) ) ;
	}
	
	return results;
}


factors_matrix	matrix ; 
size_t		sum; 
factors		used_factors ;
list<size_t>	used_numbers ;

size_t		smallest_sum ;
list<size_t>	smallest_used_numbers;

vector<size_t>  max_allowed_prime_numbers ;
bool		used_factor_two;

/*
 * returns <count> prime numbers larger than <min_value>
 */
vector<size_t> find_following_primes ( const size_t min_value, const size_t count )
{
	vector<size_t> results;	

	for ( size_t index = min_value+1; index < max_value ; ++index ) {
		if (matrix[index].count()==1 && (matrix[index].test(index))) {
			results.push_back(index);

			if (results.size() == count)
				break;
		}
	}

	return results;
}

void find_coprimes ( const size_t T, const size_t current_W )
{
	size_t delta = (used_factor_two)?2:1;
	size_t start_from = T+1;

	if ( used_factor_two && (start_from%2)==0)
		start_from++;

	if ( current_W<=0 ) {
		if ( sum < smallest_sum ) {
			smallest_sum = sum ;
			smallest_used_numbers = used_numbers ;
		}
		return ;
	}
	
	for ( size_t index = start_from ; index <= max_allowed_prime_numbers[current_W-1] ; index += delta ) {

		//check if we can use this value - 
		//if it has no common factors with the currently used factors
		factors common = used_factors & matrix[index];
		if ( common.any() )
			continue ;

		//Add this value, and dive into recursion
		used_numbers.push_back ( index ) ;
		used_factors |= matrix[index];
		sum += index ;
		if ( (index % 2) == 0 )
			used_factor_two = true;

		find_coprimes ( T, current_W-1 ) ;

		//back from recursion - remove this value
		if ( (index % 2) == 0 )
			used_factor_two = false;
		used_numbers.pop_back();
		used_factors &= ~(matrix[index]);
		sum -= index ;
	}
}

struct adder : public unary_function<double, void>
{
	adder() : sum(0) {}
	double sum;
	void operator()(double x) { sum += x; }
};


void find_sudoku_numbers ( const size_t T, const size_t W )
{
	matrix = build_factors_matrix ( max_value ) ;
	sum = 0 ;
	smallest_sum = (size_t)pow ((double) T,(double) W ) ;
	used_factors.reset();
	used_numbers.clear();
	used_factor_two = false;

	max_allowed_prime_numbers = find_following_primes ( T, W ) ;
	if ( max_allowed_prime_numbers.size() != W )
		errx(1,"Failed to find %zu prime numbers " \
			"larger than %zu (increase 'max_value'" \
			"and recompile to search above %zu)",
			W, T, max_value ) ;

	adder max_prime_sum = for_each ( max_allowed_prime_numbers.begin(),
			max_allowed_prime_numbers.end(), adder());

	find_coprimes ( T, W ) ;

	smallest_used_numbers.sort();

	cout << T << "\t"
	     << W << "\t" 
	     << max_prime_sum.sum << "\t"
	     << join_collection ( max_allowed_prime_numbers,"," ) << "\t"
	     << smallest_sum << "\t"
	     << join_collection ( smallest_used_numbers, "," )
	     << endl;
}

void show_factors ( const factors& f ) 
{
	if (f.none())
		return ;

	bool first = true ;
	for ( size_t i = 0 ; i < f.size() ; ++i ) {
		if ( f.test(i) ) {
			if ( first ) 
				first = false;
			else
				cout << " " ;

			cout << i;
		}
	}
	cout << endl;
}

void show_factors ( const size_t value )
{
	factors f = make_factorization_bitset ( value ) ;
	f.set(1);
	show_factors ( f ) ;
}

factors calc_all_common_factors ( const vector<size_t>& values ) 
{
	if (values.size()==0)
		errx(1,"values vector is empty!");

	factors common = make_factorization_bitset(values[0]);

	for ( size_t i = 1 ; i < values.size(); ++i ) {
		common &= make_factorization_bitset ( values[i] ) ;
	}

	return common ;
}

factors calc_any_common_factors ( const vector<size_t>& values ) 
{
	if (values.size()==0)
		errx(1,"values vector is empty!");

	factors common_any ;

	for ( size_t i = 0 ; i < values.size(); ++i ) {
		for ( size_t j = 0; j < values.size() ; ++j ) {
			if ( i==j )
				continue ;

			factors pair_common = 
				make_factorization_bitset ( values[i] ) 
				&
				make_factorization_bitset ( values[j] );

			if ( pair_common.any() )
				common_any |= pair_common ;
		}
	}

	return common_any ;
}

void show_matrix ( const size_t value )
{
	factors_matrix m = build_factors_matrix ( value ) ;

	cout << "-\t"; 
	for ( size_t i = 0 ; i <= value ; i++ ) {
		if ( i%10==0 )
			cout << (i/10) ;
		else
			cout << " " ;
	}
	cout << endl;

	for ( size_t i = 0 ; i <= value ; ++i ) {
		factors f = m[i];
		cout << i ;
		if ( f.count()==1 )
			cout << "p" ;
		cout << "\t" ;

		for ( size_t b = 0 ; b <= value ; ++b ) 
			cout << f[b] ;

		cout << endl;
	}
}


void usage()
{
	cout <<
"usage: find_coprime [-h] [-q] [-f N] [-c N1,N2,N3..] [-C N1,N2...] [-m N] [-s T,W]\n" \
"by A. Gordon (gordon@cshl.edu)\n" \
"\n"\
"   -h       = This helpful help screen.\n" \
"   -q       = Quiet mode - don't print anything, just return exit code.\n" \
"              Useful for scripting, see -f and -c for possible exit codes.\n" \
"   -f N     = show factors of N\n" \
"              Exit with code  0 (SUCCESS) if N is prime.\n" \
"              Exit with code -1 (FAILURE) if N is not prime.\n" \
"   -c N1... = show common factors between any two integers in the given list.\n" \
"              1 is never shown.\n" \
"              Returns empty string of no common factors found.\n" \
"              Exit with code  0 (SUCCESS) if there are NO common factors.\n" \
"              Exit with code -1 (FAILURE) if there are common factors.\n" \
"              (Make sure there's no space between the values.)\n" \
"   -C N1... = show common factors between ALL integers N1,N2,N3....\n" \
"   -m N     = Print factors bit-matrix of size N\n" \
"   -s T,W   = Find W co-primes numbers whose sum is larger than T.\n" \
"              (minimal sum)\n"\
"\n" \
"The difference between -c (any) and -C (all):\n" \
"  \"-c 5,6,9\" will return 3 (common factor between 6 and 9).\n" \
"  \"-C 5,6,9\" will return EMPTY string (no common factor between 5, 6 and 9).\n" \
"\n" \
"Scripting examples:\n" \
"  $ find_coprimes -q -f 13 && echo '13 is prime.'\n" \
"  $ find_coprimes -q -c 5,6,9 && echo '5,6,9 are co-primes.'\n" \
"\n" ;

	exit(EXIT_FAILURE);
}

vector<size_t> parse_values ( const char* optarg )
{
	vector<size_t> results;
	size_t N;
	const char* input=optarg;
	char* endptr=NULL;
	factors f;
	factors common_factors ;

	while ( *input != 0 ) {
		N = strtoul ( input, &endptr, 10 ) ;

		if ( (*endptr!=0) && ( N==0 || *endptr != ',') )
			errx(1,"Invalid value specified (%s) format should be 'N1,N2...'", optarg);
		if (N>max_value)
			errx(1,"value (%zu) too large (max. value=%zu)", N, max_value);

		results.push_back(N);

		if ( *endptr == ',' )
			endptr++;
		
		input = endptr;
	}
	return results;
}

void do_show_factors ( const char* optarg ) 
{
	vector<size_t> values = parse_values ( optarg );
	if ( values.empty() )
		errx(1,"Missing values on command line!");
	if ( values.size() != 1 )
		errx(1,"Too many values on command line!");

	if (!flag_quiet)
		show_factors ( values[0] ) ;

	exit( is_prime ( values[0] ) ? EXIT_SUCCESS : EXIT_FAILURE );
}


void do_show_all_common_factors ( const char* optarg )
{
	vector<size_t> values = parse_values ( optarg ) ;
	if (values.size()==0) 
		errx(1,"Failed to parse integer values from command line");

	factors f = calc_all_common_factors ( values ) ;
	if (!flag_quiet)
		show_factors ( f ) ;

	exit( f.none() ? EXIT_SUCCESS : EXIT_FAILURE );
}

void do_show_any_common_factors ( const char* optarg )
{
	vector<size_t> values = parse_values ( optarg ) ;
	if (values.size()==0) 
		errx(1,"Failed to parse integer values from command line");
	
	factors f = calc_any_common_factors ( values ) ;

	if (!flag_quiet)
		show_factors ( f ) ;

	exit( f.none() ? EXIT_SUCCESS : EXIT_FAILURE );
}

void do_show_matrix( const char* optarg ) 
{
	size_t N;
	char* endptr=NULL;

	N = strtoul(optarg, &endptr, 10);
	if (N==0 || *endptr != 0)
		errx(1,"Invalid value specified (%s)", optarg);

	if (N>max_value)
		errx(1,"Value too large (max. value=%zu)", max_value);

	show_matrix ( N ) ;
	exit(EXIT_SUCCESS);
}

void do_find_sudoku_numbers ( const char* optarg ) 
{
	vector<size_t> values = parse_values ( optarg ) ;
	if (values.empty()) 
		errx(1,"Failed to parse integer values from command line");
	if (values.size()!=2)
		errx(1,"Expecting 2 values on command line: T,W (got %d values)", values.size());

	find_sudoku_numbers ( values[0], values[1]) ;
	exit(0);
}

int main(int argc, char* argv[])
{
	int opt;
	while ( (opt=getopt(argc,argv,"qhf:C:c:m:s:"))!=-1 ) {
		switch(opt)
		{
		case 'q':
			flag_quiet = true;
			break;

		case 'h':
			usage();

		case 'f':
			do_show_factors ( optarg ) ;

		case 'c':
			do_show_any_common_factors ( optarg ) ;

		case 'C':
			do_show_all_common_factors ( optarg ) ;

		case 'm':
			do_show_matrix ( optarg ) ;

		case 's':
			do_find_sudoku_numbers ( optarg ) ;
		}
	}

	usage();

	return 0;
}
